/*;  khbintree.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.1 $$ {\rhead{\sl khbintree.cc\quad\thepage}} */
/*;  $$Date: 1997/09/14 09:50:41 $$  */
/*;  $$Source: /home/shiiba/KhC++LibDevelop/trees/khbintree.cc,v $$  */

/*
ΥץϡScott Robert Ladd ˤ C++ Templates and Tools
˷ǺܤƤץ򸵤ˤƤޤ

Υץϡ㳰(try, throw, catch ʤɤȤ)Υɤ
Ǥޤ̾Υ顼(顼å񤤤ơexit )
֤Ƥޤ
*/

#include <iostream>
#include "khbintree.h"
#include <assert.h>

template <class D>
    KhTreeNode<D>::KhTreeNode
        (
        const D & item,
        KhTreeNode<D> * sentinel
        )
        : Data(item)
        {
        Parent = sentinel;
        Less   = sentinel;
        More   = sentinel;
        }

// create sentinel node
template <class D>
KhTreeNode<D>::KhTreeNode()
//        : Data()
{
  Parent = this;
  Less   = this;
  More   = this;
}

/*
  Υ󥹥饯ȤǤϡScott Robert Ladd 񤤤ǤǤϡData()
  Ȥ default constructor ƤӽФ code ޤ
  դȥȤޤ code ȡint, float,
  double ʤɤ C++ Ȥߤߤηƥץ졼ȤȤ2ʬڤ
  뤳ȤǤޤ󡣥ѥΤȤˡerror message Ǥޤ

   code ̵Ƥפʤ褦Ǥ
*/

template <class D>
    KhTreeNode<D>::KhTreeNode
        (
        const KhTreeNode<D> & node
        )
        : Data(node.Data)
        {
        Parent = node.Parent;
        Less   = node.Less;
        More   = node.More;
        }

template <class D>
    void KhTreeNode<D>::operator =
        (
        const KhTreeNode<D> & node
        )
        {
        Parent = node.Parent;
        Less   = node.Less;
        More   = node.More;
        Data   = node.Data;
        }

// constructors
template <class D>
    KhBinaryTree<D>::KhBinaryTree()
        {
        Sentinel  = new KhTreeNode<D>;

		assert(Sentinel != NULL);

        Root      = Sentinel;
        LockCount = 0;
        }

template <class D>
    KhBinaryTree<D>::KhBinaryTree
        (
        const KhBinaryTree<D> & tree
        )
        {
        Sentinel = new KhTreeNode<D>;

		assert(Sentinel != NULL);

        Root      = Sentinel;
        LockCount = 0;
        RecursiveCopy(tree.Root, tree.Sentinel);
        }

template <class D>
    KhBinaryTree<D>::~KhBinaryTree()
        {
		assert(LockCount <= 0);

        RecursiveDelete(Root);

        delete Sentinel;
        }

template <class D>
    void KhBinaryTree<D>::operator =
        (
        const KhBinaryTree<D> & tree
        )
        {
		assert(LockCount <= 0);

        RecursiveDelete(Root);
        Root = Sentinel;
        RecursiveCopy(tree.Root, tree.Sentinel);
        }

// function to create node
template <class D>
    KhTreeNode<D> * KhBinaryTree<D>::CreateNode
        (
        const D & item
        )
        {
        KhTreeNode<D> * z = new KhTreeNode<D>(item, Sentinel);

		assert(z != NULL);

        return z;
        }

template <class D>
    void KhBinaryTree<D>::DeleteNode
        (
        KhTreeNode<D> * node
        )
        {
        delete node;
        }

// internal insert function
template <class D>
KhTreeNode<D> * KhBinaryTree<D>::InternalInsert
(
 const D & item
)
{
  assert(LockCount <= 0);

  KhTreeNode<D> * y = Sentinel;
  KhTreeNode<D> * x = Root;

  while (x != Sentinel)
  {
    y = x;

    if (item == x->Data)
    {
      x->Data = item;
      return NULL;
    }
    else
    {
      if (item < x->Data)
      x = x->Less;
      else
      x = x->More;
    }
  }

  KhTreeNode<D> * z = CreateNode(item);
  z->Parent = y;

  if (y == Sentinel)
  Root = z;
  else
  {
    if ((z->Data) < (y->Data))
    y->Less = z;
    else
    y->More = z;
  }

  return z;
}

// find minimum node
template <class D>
    KhTreeNode<D> * KhBinaryTree<D>::Minimum
        (
        KhTreeNode<D> * node
        )
        {
        if (node != Sentinel)
            while (node->Less != Sentinel)
                node = node->Less;

        return node;
        }

// find maximum node
template <class D>
    KhTreeNode<D> * KhBinaryTree<D>::Maximum
        (
        KhTreeNode<D> * node
        )
        {
        if (node != Sentinel)
            while (node->More != Sentinel)
                node = node->More;

        return node;
        }

// find successor node
template <class D>
KhTreeNode<D> * KhBinaryTree<D>::Successor
(
 KhTreeNode<D> * node
)
{
  KhTreeNode<D> * x, * y;

  if (node->More != Sentinel)
  return Minimum(node->More);
  else
  {
    x = node;
    y = node->Parent;

    while ((y != Sentinel) && (x == y->More))
    {
      x = y;
      y = y->Parent;
    }
  }

  return y;
}

// find predecessor node
template <class D>
    KhTreeNode<D> * KhBinaryTree<D>::Predecessor
        (
        KhTreeNode<D> * node
        )
        {
        KhTreeNode<D> * x, * y;

        if (node->Less != Sentinel)
            return Maximum(node->Less);
        else
            {
            x = node;
            y = node->Parent;

            while ((y != Sentinel) && (x == y->Less))
                {
                x = y;
                y = y->Parent;
                }
            }

        return y;
        }

// find node containing specific item
template <class D>
KhTreeNode<D> * KhBinaryTree<D>::Search
(
 const D & item
)
{
  KhTreeNode<D> * n = Root;

  while ((n != Sentinel) && !(n->Data == item))
  {
    if (item < n->Data)
      n = n->Less;
    else
      n = n->More;
  }

  return n;
}

template <class D>
void KhBinaryTree<D>::Insert
(
 const D & item
)
{
  InternalInsert(item);
}

template <class D>
KhBoolean KhBinaryTree<D>::Delete
(
 const D & item
)
{
  assert(LockCount <= 0);

  // find node
  KhTreeNode<D> * z = Search(item);

  if (z == Sentinel)
  return NO;

  KhTreeNode<D> * y, * x;

  // find node to splice out
  if ((z->Less == Sentinel) || (z->More == Sentinel))
  y = z;
  else
  y = Successor(z);

  // find child with which to replace y
  if (y->Less != Sentinel)
  x = y->Less;
  else
  x = y->More;

  // splice child onto parent
  if (x != Sentinel)
  x->Parent = y->Parent;

  if (y->Parent == Sentinel)
  Root = x; // replace root
  else
  {
    // splice in child node
    if (y == y->Parent->Less)
    y->Parent->Less = x;
    else
    y->Parent->More = x;
  }

  // if needed, save y data
  if (y != z)
  z->Data = y->Data;

  // free memory
  DeleteNode(y);

  return YES;
}

template <class D>
    void KhBinaryTree<D>::Walk
        (
        void (* func)(const D & item)
        )
        {
        WalkFunc = func;
        RecurseWalk(Root);
        }

template <class D>
    void KhBinaryTree<D>::Erase()
        {
		assert(LockCount <= 0);

        RecursiveDelete(Root);

        Root = Sentinel;
        }

template <class D>
    void KhBinaryTree<D>::RecursiveCopy
        (
        KhTreeNode<D> * node,
        KhTreeNode<D> * sent
        )
        {
        if (node != sent)
            {
            Insert(node->Data);
            RecursiveCopy(node->Less,sent);
            RecursiveCopy(node->More,sent);
            }
        }

template <class D>
    void KhBinaryTree<D>::RecurseWalk
        (
        KhTreeNode<D> * node
        )
        {
        if (node != Sentinel)
            {
            RecurseWalk(node->Less);
            WalkFunc(node->Data);
            RecurseWalk(node->More);
            }
        }

template <class D>
    void KhBinaryTree<D>::RecursiveDelete
        (
        KhTreeNode<D> * node
        )
        {
        if (node != Sentinel)
            {
            RecursiveDelete(node->Less);
            RecursiveDelete(node->More);
            DeleteNode(node);
            }
        }

// manipulate lock count
// modified by Yutaka ICHIKAWA (Tue Jul 13 21:03:24 1999)
// template <class D>
//     inline size_t KhBinaryTree<D>::GetLockCount()
template <class D>
    inline unsigned int KhBinaryTree<D>::GetLockCount()
        {
        return LockCount;
        }

template <class D>
    inline const KhTreeNode<D> * KhBinaryTree<D>::GetSentinel()
        {
        return Sentinel;
        }

template <class D>
D* KhBinaryTree<D>::GetPtrToData(const D & item)
{
  KhTreeNode<D>* x = Search(item);
  if (x == Sentinel)
    return NULL;
  else
    return &(x->Data);
}

template <class D>
KhBinaryTreeIterator<D>::KhBinaryTreeIterator
(
 KhBinaryTree<D> & bt
)
  : Tree(bt)
{
  assert(Tree.LockCount < UINT_MAX);

  ++Tree.LockCount;

  Smallest();
}

template <class D>
    KhBinaryTreeIterator<D>::KhBinaryTreeIterator
        (
        KhBinaryTreeIterator<D> & iter
        )
        : Tree(iter.Tree)
        {
		assert(Tree.LockCount < UINT_MAX);

        ++Tree.LockCount;

        Node = iter.Node;
        }

template <class D>
    KhBinaryTreeIterator<D>::~KhBinaryTreeIterator()
        {
		assert(Tree.LockCount > 0);

        --Tree.LockCount;
        }

template <class D>
    void KhBinaryTreeIterator<D>::operator =
        (
        KhBinaryTreeIterator<D> & iter
        )
        {
        Tree = iter.Tree;

		assert(Tree.LockCount > 0);

        --Tree.LockCount;

		assert(Tree.LockCount < UINT_MAX);

        ++Tree.LockCount;

        Node = iter.Node;
        }

template <class D>
    void KhBinaryTreeIterator<D>::Smallest()
        {
        Node = Tree.Minimum(Tree.Root);
        }

template <class D>
    void KhBinaryTreeIterator<D>::Largest()
        {
        Node = Tree.Maximum(Tree.Root);
        }

template <class D>
    void KhBinaryTreeIterator<D>::operator ++ ()
        {
        Node = Tree.Successor(Node);
        }

template <class D>
    void KhBinaryTreeIterator<D>::operator -- ()
        {
        Node = Tree.Predecessor(Node);
        }

template <class D>
D KhBinaryTreeIterator<D>::operator * ()
{
  assert(Node != Tree.GetSentinel());
  return Node->Data;
}

template <class D>
KhBinaryTreeIterator<D>::operator D* ()
{
  if (Node == Tree.GetSentinel())
    return NULL;
  else
    return &(Node->Data);
}

//template <class D>
//KhBinaryTreeIterator<D>::operator void*()
//{
//  if (Node == Tree.GetSentinel())
//    return NULL;
//  else
//    return this;
//}

/*; end of file */
