/*;  khmatrix.h {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.1 $$ {\rhead{\sl khmatrix.h\quad\thepage}} */
/*;  $$Date: 1998/07/01 11:51:03 $$  */
/*;  $$Source: /home/shiiba/KhProgDevelop/KhC++Lib/matrix/khmatrix.h,v $$  */
/*;  $$RCShost$$  */

//;=============================================================
//;  Matrix Classes     khmatrix.h   v2.00
//;      Template for matrices
//;      Copyright 1993-1995 Scott Robert Ladd
//;=============================================================

#ifndef MSMATRIX_H
#define MSMATRIX_H

#include <stddef.h>
#include <limits.h>
#include <math.h>
#include <string.h>	// memset ϡƤޤ
#include <iomanip> // mod Tue Aug  8 16:39:18 2006

using namespace std; // added Tue Aug  8 16:39:25 2006

#include "khgrid.h"
#include "khboolean.h"
#include "khsimple.h"
#include "kherror.h"

//;-------------------------------------------------
//;  Matrix
//;      A two-dimensional mathematical matrix class
//;-------------------------------------------------

template <class D>
class KhMatrix : public KhGrid<D>
{
 public:
  //; constructors
  KhMatrix();

  KhMatrix		 // 󥹥ȥ饯
  (
   size_t rows,		 // Կ
   size_t cols,		 // 
   const D & init = (D)0 // Ǥν
   );

  KhMatrix
  (
   const KhMatrix<D> & m
   );

  KhMatrix
  (
   const KhGrid<D> & m
   );

  // added by iYutaka (Fri Oct 28 15:11:05 2005)
  KhMatrix
  (
   D x
   );

  // assignment operator
  void operator =
  (
   const KhMatrix<D> & m
   );

  void operator =
  (
   const KhGrid<D> & m
   );

  void operator =	// εˤΤޤʣ
  (
   const D * array
   );

  // interrogation
  KhBoolean IsZero() const;
  KhBoolean IsDiagonal() const;
  KhBoolean IsIdentity() const;
  KhBoolean IsTridiagonal() const;
  KhBoolean IsUpperTriangular() const;
  KhBoolean IsLowerTriangular() const;
  KhBoolean IsPermutation() const;

  // scalar addition
  KhMatrix<D> operator + ( // äΤͤȤ֤
    const KhMatrix<D> & m
    );

  void operator += (	 // ˹­ࡣʬѤ롣
    const KhMatrix<D> & m
    );

  KhMatrix<D> operator + ( // ƤʬƱο x ­Τ
			 // ֤
    const D & x
    );

  void operator += (	 // ƤʬƱο x ­
    const D & x
    );

  // scalar subtraction
  KhMatrix<D> operator - ( // ΤͤȤ֤
    const KhMatrix<D> & m
    );

  void operator -= (       // ʬѤ
    const KhMatrix<D> & m
    );

  KhMatrix<D> operator - ( // ƤʬƱο x 
			   // ͤȤ֤
    const D & x
    );

  void operator -= (	 // ƤʬƱο x ʬѤ
			 // 롣
    const D & x
    );

  // scalar multiplication
  KhMatrix<D> operator * ( // ʬƱΤѤȤäƺäͤȤ֤
    const KhMatrix<D> & m
    );

  void operator *=	 // ʬƱΤѤȤäƺäѹ롣
  (
   const KhMatrix<D> & m
   );

  KhMatrix<D> operator * ( // ƤʬƱο x ݤͤȤ
			 // ֤
    const D & x
    );

  void operator *= (	 // ƤʬƱο x ݤ
    const D & x
    );

  // scalar division
  KhMatrix<D> operator / (
    const KhMatrix<D> & m
    );

  void operator /=
  (
   const KhMatrix<D> & m
   );

  KhMatrix<D> operator /
  (
   const D & x
   );

  void operator /=
  (
   const D & x
   );

  // matrix multiplication
  KhMatrix<D> operator %  // ѡ̤ͤȤ֤
  (
   const KhMatrix<D> & m	// ݤ
   );

  // comparison operators
  KhBoolean Equals        // 
  (
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator == ( // ʬƱΤ̤Ǽ
			    // ֤
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator !=
  (
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator <
  (
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator <=
  (
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator >
  (
   const KhMatrix<D> & m
   );

  KhMatrix<int> operator >=
  (
   const KhMatrix<D> & m
   );

  // negate a matrix
  KhMatrix<D> operator - ();

  // change size (destroying contents)
  void Resize
  (
   size_t rows,
   size_t cols,
   const D & init = D(0)
   );

  // inner and outer products

  D InnerProduct	// ԥ٥ȥƱΤޤ٥ȥƱΤ
  (
   const KhMatrix<D> & m
   ) const;

  // calculation euclidean norm
  double Norm() const;

  // calculate determinant value
  D Determinant() const;
                        // 󼰤͡󼰤ŸȤäơ
			// 󼰤ͤƵŪ˷׻Ƥ롣

  // create a minor matrix

  KhMatrix<D> Minor
  (
   size_t rdel,
   size_t cdel
   ) const;

  // rdel Ԥ cdel ä֤ޤ

  // solve system of linear equations

  KhBoolean LinSolve(KhMatrix<D>& cv);

  // õˡˤä켡 A % x = b 򤭤ޤϡ
  // R*(R+1) Ǥꡤǽ R ˷ A  R+1  b
  // ǼƤΤȤơޤβǡѹ
  // ޤ򡤰ȤͿ줿 R*1 (٥ȥ) cv λȤ
  // ޤ

  // §Ǥʤˡ˲򤬵ʤäϡ
  // ͤ NO Ǥξ硤ȤͿ줿 cv ͤѹޤ
  // ϡѹƤޤ˲򤬵ޤäͤ
  // YES Ǥ

 protected:
  // internal recursive function for determinant
  D DetRecursive() const;
};

// constructors
template <class D>
inline KhMatrix<D>::KhMatrix()
  : KhGrid<D>()
{
  *KhGrid<D>::Data = D(0);
}

template <class D>
inline KhMatrix<D>::KhMatrix
(
 const KhMatrix<D> & m
 )
  : KhGrid<D>(m)
{
  // place holder
}

template <class D>
inline KhMatrix<D>::KhMatrix
(
 const KhGrid<D> & m
 )
 : KhGrid<D>(m)
{
  // place holder
}

// assignment operator
template <class D>
inline void KhMatrix<D>::operator =
(
 const KhMatrix<D> & m
 )
{
  KhGrid<D>::operator = (m);
}

template <class D>
inline void KhMatrix<D>::operator =
(
 const KhGrid<D> & m
 )
{
  KhGrid<D>::operator = (m);
}

template <class D>
inline void KhMatrix<D>::operator =
(
 const D * array
 )
{
  KhGrid<D>::operator = (array);
}

template <class D>
ostream& operator<<(ostream& s, KhMatrix<D>& a);

template <class D>
istream& operator>>(istream& s, KhMatrix<D>& a);

/*

------------------------------------------------------------------------
饹 Lufact<Type>  LU ʬ갷Υ饹Ǥ

LU ʬȤϡ§ʹ A Ϳ줿Ȥˡгʬ٤ 1 Ǥ
3ѹ L ȡ3ѹ U Ѥη

  A = LU

ʬ򤹤뤳Ȥ򤤤ޤ

¤ϡ A §ǤäƤ⡤ɬ⤳Τ褦ʷʬǤȤϸ¤
ޤ󡣤 A §ǤСŬ˹Ԥν֤ѹС
ʬǤޤäơ LU ʬ򰷤ˤϡԤν֤ɤΤ
ѹ뤫ȤˤĤɬפޤ

Υإåե롤LU ʬ갷Υ饹ǤϡLU ʬ
̤ L  UԤν֤ɤΤ褦ѹʤɤξޤȤƥǡ
ФȤƵ뤳ȤˤдؿȤƤޤ

------------------------------------------------------------------------
* 饹 Lufact<Type> Υǡ

饹 Lufact ΥǡФϡ̤Ǥ

 private:
  KhMatrix<Lufact> alu;   // LUʬ򤷤
  KhMatrix<int> indx;     // LUʬκݤιԸ򴹤εϿ
  Type d;               // 1 ޤ -1 ޤ 0

ǡType ϡfloat ޤ double Ǥ

LU ʬϥ󥹥ȥ饯¹Ԥޤ󥹥ȥ饯ϡ򤦤Ȥ
 LU ʬ¹Ԥη̤ǡ alu ȤƵޤL ϡ
гʬ 1 β3ѹǤгʬΤ3ʬalu β3
Ѥ˵ޤU Ͼ3ѹǤalu ξ3ʬ alu ξ壳
˵ޤ

indx ϡLU ʬβǹԤʤäԸ򴹤ηвϿƤޤindx(j) =
k ϡ j ֤ΥԥܥåȤܤȤȤˡ k ԤˤäԤ j 
˰ư k Ԥ j ԤäƤäơLU ʬ򤷤Ȥ̣ޤ
ΤȤk Ԥϡǽ餫 k ԤˤäȤϸ¤ʤȤդƤ


Ը򴹤βʤ顤d ͤ 1 ˡʤ顤-1 ˥åȤޤ

LU ʬ򤷤褦ȤƤ§ǤʤϡLU ʬǻߤᡤd 
0.0 ˥åȤޤξ硤alu  indx ˤޤǤη̤äƤ
Τޤ󤬡ͤˤϰ̣Ϥޤ

------------------------------------------------------------------------
* 饹 Lufact<Type> Υдؿ

Lufact(const Matirx<Type>& A)

  LU ʬ򤷤褦Ȥ A.

KhBoolean IsRegular()

  LU ʬ򤷤褦Ȥ§ʤ YES ðۤʤ NO 
  ޤ

KhBoolean IsSingular()

  LU ʬ򤷤褦Ȥðۤʤ YES §ʤ NO 
  ޤ

Type Determinant()

  LU ʬ򤷤褦Ȥι󼰤֤ͤޤ

void Solve(KhMatrix<Type>& b)

  A x = b 򤭤ޤϡ٥ȥ b ֤ޤb ιԿ
    A ιԿƱǤʤˤϡ顼å񤤤ơ
  exit ޤ

KhBoolean Invmatrix(KhMatrix<Type>& invm)

  LU ʬ򤷤褦Ȥεչ֤ޤLU ʬ򤷤
  Ȥ§ǤʤˤϡNO ֤ޤ

*/

template <class Type>
class Lufact
{
 private:
  KhMatrix<Type> Alu;
  KhMatrix<int> Indx;
  Type D;

 public:
  Lufact(const KhMatrix<Type>& a);
  ~Lufact() {}
  KhBoolean IsRegular();
  KhBoolean IsSingular();
  Type Determinant();
  void Solve(KhMatrix<Type>& b);
  KhBoolean Invmatrix(KhMatrix<Type>& invm);
};

#endif

/*; end of file */
