// kinem.cc
// modified by T.Sayama on Jan. 17, 2005.
// ver 2.3 ԗʂvZJn݂̂Ƃʂ`}ꍇƁC
//                     efʂ̗ʂǂݍޏꍇ̗ɑΉD
//         ҂Ŏgpꍇ́Clt@CɂāCvZJn̉
//         mDivLqC̍smDiv+1̊efʂ̏ʂLqD


#include <assert.h>
#include <math.h>
#include "kinem.h"   // NX Kinem ̒`

double max(double a, double b)
{
    return (a >= b) ? a : b ;
}

Kinem::Kinem
(
	char* o_name, 
	int o_num, 
	int aNumUpInflow, 
	int aNumLateral
)
: Element()
{
	class_name = "Kinem";
	Init(o_name, o_num, aNumUpInflow, aNumLateral);
}

Kinem::Kinem(void) : Element()
{
	class_name = "Kinem";
}

Boolean Kinem::Init
(
	char* o_name, 
	int o_num,
    int aNumUpInflow, 
 	int aNumLateral
)
{
	if (aNumUpInflow < 0)
	{
		Eprint(__FILE__, __LINE__);
		    // Kinem::Init
			// Number of inflows from upper end must not be negative number.
		Oprint();
		exit(1);
	}
	mNumUpInflow = aNumUpInflow;

	if ((rpu = new Receive<double>[mNumUpInflow]) == NULL)
	{
		Eprint(__FILE__, __LINE__);
		    // Kinem::Init
			// Cannot allocate memory for rpu.
		Oprint();
		exit(1);
	}

	if (aNumLateral < 0) {
		Eprint(__FILE__, __LINE__);
		    // Kinem::Init
			// Number of lateral inflows must not be negative number.
		Oprint();
		exit(1);
	}
	mNumLateral = aNumLateral;

	if ((rpl = new Receive<double>[mNumLateral]) == NULL) {
		Eprint(__FILE__, __LINE__);
		    // Kinem::Init
			// Cannot allocate memory for rpl.
		Oprint();
		exit(1);
	}

	for (int i = 0; i < mNumUpInflow; i++) rpu[i].Init("rpu", i);
	for (int i = 0; i < mNumLateral; i++) rpl[i].Init("rpl", i);

	sp.Init("sp", 0);

	return Element::Init(o_name, o_num, 0);
}

void Kinem::Register_receive_ports(void)
{
	for (int i = 0; i < mNumUpInflow; i++) Register((void*)&rpu[i]);
	for (int i = 0; i < mNumLateral; i++) Register((void*)&rpl[i]);
}

void Kinem::Set_parameter(FILE* fp)
{
	if (fscanf(fp, "%lf %lf %lf %lf %d %ld %lf", 
			   &mL, &mB, &mG, &mMan, &mDiv, &mDt, &mTheta) != 7)
	{
		Eprint(__FILE__, __LINE__);
		    // Kinem::Set_parameter
			// Cannot read parameter values.
		Oprint();
		exit(1);
	}

	assert(mL > 0.0);
	assert(mB > 0.0);
	assert(mG > 0.0);
	assert(mMan > 0.0);
	assert(mDiv > 0);
	assert(mDt > 0);
	assert(0.5 < mTheta && mTheta <= 1.0);

	mDx = mL / mDiv;
	mG = atan(mG); mG = sin(mG); // sin ̒lɕϊ

	mQc = new double[mDiv+1];
	mQn = new double[mDiv+1];
	mAc = new double[mDiv+1];
	mAn = new double[mDiv+1];
	mK = new double[mDiv+1];
	mZ = new double[mDiv+1];
	mAlpha = new double[mDiv+1];
	mM = new double[mDiv+1];

	for (int i = 0; i <= mDiv; i++)	fscanf(fp, "%lf", &mK[i]);
	for (int i = 0; i <= mDiv; i++)	fscanf(fp, "%lf", &mZ[i]);

	for (int i = 0; i <= mDiv; i++)	
	{
		mAlpha[i] = 1.0 / mMan * pow(mK[i], 0.666667) * sqrt(mG);
		mM[i] = 1.0 + 0.666667 * mZ[i];
	}
}

// void Kinem::Set_initial_state(FILE* fp)
// {
// 	// vZJn
// 	// ͓[̗ (m3/s)
// 	// ͓[̗o (m3/s)
// 	//  (m3/s)
// 
// 	assert(fscanf(fp, "%ld %lf %lf %lf", 
// 				  &current_time, &mQc[0], &mQc[mDiv], &mQlc) == 4);
// 
// 	assert(mQc[0] >= 0.0);
// 	assert(mQc[mDiv] >= 0.0);
// 	assert(mQlc >= 0.0);
// 
// 	mQlc /= mL;	// Pʕɕϊ
// 
// 	// ʂ̐`Ԃђʐfʐςւ̕ϊ
// 	for (int i = 0; i <= mDiv; i++)
// 	{
// 		mQc[i] = mQc[0] + (mQc[mDiv] - mQc[0]) * i / mDiv;
// 		mAc[i] = pow(mQc[i]/mAlpha[i], 1.0/mM[i]);
// 	}
// }

void Kinem::Set_initial_state(FILE* fp)
{
	// vZJnǂݍށB
	//assert(fscanf(fp, "%ld", &current_time) == 1);
	char buf[256];
	int mDiv_temp;

	// 󔒍s̓ǂݔ΂
	while(fgets(buf, 256, fp) != NULL & sscanf(buf, "%ld", &current_time) != 1){}

	if(sscanf(buf, "%ld %d", &current_time, &mDiv_temp) == 2)
	{
		// ԗʂǂݍ߂ꍇ
		initFlag = 1;
		if(mDiv_temp != mDiv)
		{
			Eprint(__FILE__, __LINE__);
			// Kinem::Set_initial_state
			// Cannot read initial state file.
			Oprint();
			exit(1);
		}
		for (int i = 0; i <= mDiv; i++)	fscanf(fp, "%lf", &mQc[i]);
	}
	else if(sscanf(buf, "%ld", &current_time) == 1)
	{
		// ԗʂǂݍ߂Ȃꍇ
		initFlag = 0;
	}
	else
	{
		Eprint(__FILE__, __LINE__);
	    // Kinem::Set_initial_state
		// Cannot read initial state file.
		Oprint();
		exit(1);
	}
}

void Kinem::Save_terminal_state(FILE* fp)
{
	assert(fp != NULL);

	//fprintf(fp, "%ld %f %f %f\n", 
	//	current_time, mQc[0], mQc[mDiv], mQlc);

	// ԗʂǂݍޏꍇ
	//if(initFlag == 1){
	fprintf(fp, "%ld %d\n", 
		current_time, mDiv);
	for (int i = 0; i <= mDiv; i++)	fprintf(fp, "%f ", mQn[i]);
	fprintf(fp, "\n");
	//}
	// ԗʂǂݍ܂Ȃꍇ
	//else{
	//	fprintf(fp, "%ld\n", current_time);
	//}
}

time_t Kinem::Calculate_time_step(void)
{
	return mDt;
}

Boolean Kinem::Initial_output(void)
{
	// vZJn̏[̗ʁAʂM[q擾A
	// ͓̏ʂvZƂƂɁA[̗oʂ𑗐M[q
	// oB

	double wQ = 0.0;
	for (int i = 0; i < mNumUpInflow; i++)
	{
		if (rpu[i].Can_you_get_data(current_time, current_time) == NO)
		{
			return NO;
		}
		wQ += rpu[i].Get_data(current_time);
	}

	double wQl = 0.0;
	for (int i = 0; i < mNumLateral; i++)
	{
		if (rpl[i].Can_you_get_data(current_time, current_time) == NO)
		{
			return NO;
		}
		wQl += rpl[i].Get_data(current_time);
	}

	// ԗʂǂݍޏꍇ
	if(initFlag == 1)
	{
		mQlc = wQl / mL;	// Pʕɕϊ

		// ʂ̒ʐfʐςւ̕ϊ
		for (int i = 0; i <= mDiv; i++)
		{
			mAc[i] = pow(mQc[i]/mAlpha[i], 1.0/mM[i]);
		}
	}
	// ԗʂǂݍ܂Ȃꍇ
	else{
		mQc[0] = wQ;
		mQc[mDiv] = wQ + wQl;

		mQlc = wQl / mL;	// Pʕɕϊ

		// ʂ̐`Ԃђʐfʐςւ̕ϊ
		for (int i = 0; i <= mDiv; i++)
		{
			mQc[i] = mQc[0] + (mQc[mDiv] - mQc[0]) * i / mDiv;
			mAc[i] = pow(mQc[i]/mAlpha[i], 1.0/mM[i]);
		}
	}

    //sp.Set_comment_string("time(sec) discharge(m**3/sec)");
	sp.Send_data(current_time, mQc[mDiv]);
	return YES;
}

Boolean Kinem::Calculate(void)
{

	double dt;
	dt = time_step;

	//  (m2/s)

	mQln = 0.0;
	for (int i = 0; i < mNumLateral; i++)
	{
		mQln += rpl[i].Get_data(current_time + time_step);
	}
	mQln /= mL;

	double wQl = 0.5 * (mQlc + mQln);

	// 㗬[̗ (m3/s)

	double wUpInflow = 0.0;
	for (int i = 0; i < mNumUpInflow; i++)
	{
		wUpInflow += rpu[i].Get_data(current_time + time_step);
	}

	// modified by T.Sayama
	if(wUpInflow < 0)
	{
		wUpInflow = 0;
	}

	mQn[0] = wUpInflow;
	mAn[0] = pow(mQn[0]/mAlpha[0], 1.0/mM[0]);

	for (int i = 1; i <= mDiv; i++)
	{
		double wCa = mAlpha[i-1]*mM[i-1]*pow(mAc[i-1], mM[i-1]-1.0);
                // double wCb = mAlpha[i]*mM[i]*pow(mAc[i], mM[i]-1.0);

                double wQ = max(mQc[i], 0.5*(mQc[i-1]+mQn[i-1])+wQl*mDx);
                double wCb = 
                    pow(mAlpha[i], 1.0/mM[i]) * mM[i] * pow(wQ, (mM[i]-1.0)/mM[i]);

// 		// ȉ̓s pow gȂ̂ňꌩǂɎv邪A
// 		// mAc[i,i-1]  0 ̂Ƃ floating exception N̂ł
// 		double wCa = mQc[i-1]/mAc[i-1]*mM[i-1];
// 		double wCb = mQc[i]/mAc[i]*mM[i];

// 		double wC = 0.5 * (wCa + wCb);
//                 if (wC <= EPSQ) wC = EPSQ;

		double wC = 0.5 * (wCa + wCb);

		mQn[i] = (mQc[i] + mTheta * wC * (dt/mDx * mQn[i-1] + dt * wQl) - 
			(1. - mTheta) * wC * dt * ((mQc[i] - mQc[i-1])/mDx - wQl)) / 
			(1. + mTheta * wC * dt / mDx);

		if (mQn[i] < 0.)
		{
			Wprint(__FILE__, __LINE__);
			// Negative discharge
			// fprintf(stderr, "mQn[i] : %f\n", mQn[i]);
			// test
			printf("[%s %d] %ld, mQln : %f, wUpInflow : %f, mQn : %f\n", object_name, object_number,
				current_time + time_step, mQln, wUpInflow, mQn[i]);
			mQn[i] = 0.;
		}

		mAn[i] = pow(mQn[i]/mAlpha[i], 1.0/mM[i]);
	}
/*
	// o (600 bɏo͂ꍇ)
	double fmod600 = fmod((double)(current_time + time_step), 600);
	if(fmod600 == 0)
	{
		sp.Send_data(current_time + time_step, mQn[mDiv]);
	}
*/	
	sp.Send_data(current_time + time_step, mQn[mDiv]);
	
	// ԗʂ̍XV
	double* c;
	c = mQc; mQc = mQn; mQn = c;
	c = mAc; mAc = mAn; mAn = c;
	mQlc = mQln;

	//printf("[%s %d] %ld : %f\n", object_name, object_number,
	//	current_time + time_step, mQn[mDiv]);

	return YES;
}
// end of file
